# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gradient_ascent']

package_data = \
{'': ['*']}

install_requires = \
['einops', 'torch']

setup_kwargs = {
    'name': 'gradient-ascent',
    'version': '0.0.2',
    'description': 'Gradient Ascent - Pytorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n# Gradient Ascent\nGradient Ascent is just the opposite of Gradient Descent. While Gradient Descent adjusts the parameters in the opposite direction of the gradient to minimize a loss function, Gradient Ascent adjusts the parameters in the direction of the gradient to maximize some objective function.\n\n\nI got the idea for this while playing basketball, I don\'t know why or how but this is my attempt to implement it.\n\n\n\n# Appreciation\n* Lucidrains\n* Agorians\n\n# Install\n`pip install gradient-ascent`\n\n# Usage\n```python\nimport torch\nfrom gradient_ascent.main import GradientAscent\n\n\nclass SimpleModel(torch.nn.Module):\n    def __init__(self):\n        super(SimpleModel, self).__init__()\n        self.fc = torch.nn.Linear(1, 1)\n\n    def forward(self, x):\n        return self.fc(x)\n\n\n# Test the optimizer\nmodel = SimpleModel()\noptimizer = GradientAscent(model.parameters(), lr=0.01)\n\n# General some sample data\ndata = torch.tensor([[2.0]])\ntarget = torch.tensor([[3.0]])\n\nfor _ in range(1000):\n    optimizer.zero_grad()\n    output = model(data)\n\n    # Negative loss as we are maximizing]\n    loss = -torch.nn.functional.mse_loss(output, target)\n    loss.backward()\n    optimizer.step()\n\nprint("Final output after training: ", model(data))\n\n```\n\n# Architecture\n\n### Theoretical Overview\nFor a function \\( f(\\theta) \\), the update step in gradient ascent is given by:\n\n\\[ \\theta_{new} = \\theta_{old} + \\alpha \\nabla f(\\theta_{old}) \\]\n\nWhere:\n- \\( \\theta \\) are the parameters.\n- \\( \\alpha \\) is the learning rate.\n- \\( \\nabla f(\\theta_{old}) \\) is the gradient of the function with respect to the parameters.\n\n### Gradient Ascent Pseudocode\n\n```\nAlgorithm: GradientAscentOptimizer\n\n1. Input: \n   - Objective function f(θ)\n   - Initial parameters θ₀\n   - Learning rate α\n   - Maximum iterations max_iter\n\n2. For iteration = 1 to max_iter:\n   a. Compute gradient: ∇θ = gradient of f(θ) w.r.t θ\n   b. Update parameters: θ = θ + α * ∇θ\n\n3. Return final parameters θ\n```\n\n### 3 New Features\n1. Non-Convexity:\nMany problems in deep learning involve non-convex optimization landscapes. Gradient ascent, like gradient descent, can get stuck in local maxima when dealing with such landscapes. Adding mechanisms to escape from these local optima can be necessary.\n\n2. Momentum:\nMomentum can be integrated to accelerate gradient vectors in any consistent direction, which can help in faster convergence and also in avoiding getting stuck in shallow local maxima.\n\n3. Adaptive Learning Rates:\nThe learning rate might need to adapt based on the recent history of gradients, allowing the optimization to move faster during the early stages and slow down during fine-tuning. This is seen in optimizers like AdaGrad, RMSProp, and Adam.\n\n\n\n# Applications:\nThe Gradient Ascent with features like momentum and adaptive learning rates, as discussed, is tailored to handle challenges in non-convex optimization landscapes. Here are some tasks and scenarios where this optimizer would be particularly beneficial:\n\n1. **Maximizing Likelihoods**: \n   - Many models in machine learning are framed as maximum likelihood estimation (MLE) problems, where the goal is to adjust parameters to maximize the likelihood of the observed data. Examples include Gaussian Mixture Models, Hidden Markov Models, and some Neural Network configurations. Gradient ascent is directly applicable in such cases.\n\n2. **Generative Adversarial Networks (GANs)**:\n   - The generator in a GAN tries to maximize an objective function where it fools the discriminator. While traditional GANs use gradient descent with a flipped objective, using gradient ascent can be a more direct way to express this maximization problem.\n\n3. **Game Theoretic Frameworks**:\n   - In scenarios where multiple agents are in competition or cooperation, and their objectives are to maximize certain rewards, gradient ascent can be used. This applies to multi-agent reinforcement learning and certain types of equilibrium-seeking networks.\n\n4. **Policy Gradient Methods in Reinforcement Learning**:\n   - Policy gradient methods aim to maximize the expected reward by adjusting a policy in the direction that increases expected returns. Gradient ascent is a natural fit for this optimization problem.\n\n5. **Eigenproblems**:\n   - In tasks where the goal is to find the maximum eigenvalue of a matrix or the corresponding eigenvector, gradient ascent techniques can be applied.\n\n6. **Feature Extraction and Representation Learning**:\n   - When the goal is to learn features that maximize the variance or mutual information (e.g., Principal Component Analysis or some Information Maximization approaches), gradient ascent can be used to optimize the objective directly.\n\n7. **Sparse Coding**:\n   - The aim here is to find a representation that maximizes the sparsity under certain constraints. The problem can be reframed as a maximization problem solvable with gradient ascent.\n\nFor scenarios with non-convex landscapes, the features like **momentum** help escape shallow local maxima, and **adaptive learning rates** ensure efficient traversal of the optimization landscape, adapting the step sizes based on the gradient\'s recent history.\n\nHowever, while this optimizer can be effective in the above scenarios, one should always consider the specific nuances of the problem. It\'s essential to remember that no optimizer is universally the best, and empirical testing is often necessary to determine the most effective optimizer for a particular task.\n\n# Benchmarks\n`python benchmarks.py`\n\n```\nBenchmark 1: 9.999994277954102\nBenchmark 2: 1.375625112855263e-23\nBenchmark 3: -131395.9375\nBenchmark 4: -333186848.0\nBenchmark 5: -166376013824.0\nBenchmark 6: 0.31278279423713684\nBenchmark 7: [1.375625112855263e-23, 1.375625112855263e-23]\nBenchmark 8: -28.793724060058594\nBenchmark 9: 1.0\nBenchmark 10: 0.8203693628311157\n```\n\n# License\nMIT\n\n\n\n# Todo\n- Provide metric logging + make more dynamic\n- Add more benchmarks\n- Validate by training a small Hidden Markov Model or another model',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/gradient-ascent',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
